@@ -7,12 +7,12 @@ from inspect import iscoroutine, isawaitable
 
 from asgi_sessions import Session
 from asgi_tools.response import parse_response
-from muffin import Application, ResponseRedirect, Response, Request
+from muffin import Application, ResponseRedirect, Response, Request, ResponseError
 from muffin.plugins import BasePlugin
 from muffin.typing import Receive, Send, ASGIApp
 
 
-__version__ = "1.0.0"
+__version__ = "1.2.0"
 __project__ = "muffin-session"
 __author__ = "Kirill Klenov <horneds@gmail.com>"
 __license__ = "MIT"
@@ -73,6 +73,11 @@ class Plugin(BasePlugin):
 
         return response
 
+    def user_loader(self, func: F) -> F:
+        """Register a function as user loader."""
+        self._user_loader = func  # noqa
+        return func
+
     def load_from_request(self, request: Request) -> Session:
         """Load a session from the request."""
         if SESSION_KEY not in request:
@@ -93,11 +98,6 @@ class Plugin(BasePlugin):
         response.headers['Set-Cookie'] = obj.cookie(self.cfg.cookie_name, self.cfg.cookie_params)
         return response
 
-    def user_loader(self, func: F) -> F:
-        """Register a function as user loader."""
-        self._user_loader = func  # noqa
-        return func
-
     async def load_user(self, request):
         """Load user from request."""
         if USER_KEY not in request:
@@ -112,13 +112,15 @@ class Plugin(BasePlugin):
 
         return request[USER_KEY]
 
-    def user_pass(self, func=None, location=None, **rkwargs):
+    def user_pass(self, checker: t.Callable = None,
+                  location: t.Union[str, t.Callable[[Request], str], ResponseError] = None,
+                  **rkwargs):
         """Check that a user is logged and pass conditions."""
         def wrapper(view):
 
             @functools.wraps(view)
             async def handler(request, *args, **kwargs):
-                await self.check_user(request, func, location, **rkwargs)
+                await self.check_user(request, checker, location, **rkwargs)
                 return await view(request, *args, **kwargs)
 
             return handler
@@ -126,24 +128,28 @@ class Plugin(BasePlugin):
         return wrapper
 
     async def check_user(
-            self, request: Request, func: t.Callable = None,
+            self, request: Request, checker: t.Callable = None,
             location: t.Union[str, t.Callable] = None, **response_params) -> t.Any:
-        """Check for user is logged and pass the given func.
+        """Check for user is logged and pass the given checker.
 
-        :param func: user checker function, defaults to default_user_checker
+        :param checker: user checker function, defaults to default_user_checker
         :param location: where to redirect if user is not logged in.
             May be either string (URL) or function which accepts request as argument
             and returns string URL.
         """
         user = await self.load_user(request)
-        func = func or self.cfg.default_user_checker
-        if not func(user):
-            redirect_url = location or self.cfg.login_url
-            while callable(redirect_url):
-                redirect_url = redirect_url(request)
-                while iscoroutine(redirect_url):
-                    redirect_url = await redirect_url
-            raise ResponseRedirect(redirect_url, **response_params)
+        checker = checker or self.cfg.default_user_checker
+        if not checker(user):
+            redirect = location or self.cfg.login_url
+            if isinstance(redirect, ResponseError):
+                raise redirect
+
+            if callable(redirect):
+                redirect = redirect(request)
+                if iscoroutine(redirect):
+                    redirect = await redirect
+            raise ResponseRedirect(redirect, **response_params)
+
         return user
 
     def login(self, request: Request, ident: str, *, response: t.Any = None):
